// Reporting successor states in graph-based analysis -*- c++ -*-

#ifndef GRAPHREPORTER_H_
# define GRAPHREPORTER_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include "StateReporter.h"
# include "Search.h"

/** @file GraphReporter.h
 * Interface for reporting successor states in graph-based analysis
 */

/** Reporting successor states in graph-based analysis */
class GraphReporter : public StateReporter
{
public:
  /** Constructor
   * @param compilation	the compiled model (optional)
   * @param net_	the net that is being analysed
   * @param printer_	printer object for diagnostic output
   * @param maxerrors	maximum number of allowed errors (0=infinity)
   * @param compress	flag: collapse deterministic sequences?
   * @param local	flag: suppress transient (invisible) states?
   * @param thresold	thresold for reporting state space statistics
   * @param graph	the reachability graph being generated
   * @param states	list of unprocessed states
   * @param addOld	flag: add also existing states to the state list
   */
  GraphReporter (
#ifdef EXPR_COMPILE
		 const class Compilation* compilation,
#endif // EXPR_COMPILE
		 const class Net& net_,
		 const class Printer& printer_,
		 unsigned maxerrors,
		 bool compress,
		 bool local,
		 unsigned thresold,
		 class Graph& graph,
		 class SearchList& states,
		 bool addOld);
private:
  /** Copy constructor */
  GraphReporter (const class GraphReporter& old);
  /** Assignment operator */
  class GraphReporter& operator= (const class GraphReporter& other);
public:
  /** Destructor */
  ~GraphReporter ();

  /** Get a reference to the reachability graph */
  const class Graph& getGraph () const { return myGraph; }

  /** Retrieve the successors of a state
   * @param state	the state to be analyzed
   * @return		the successors of the state (must be cleared by caller)
   */
  class SearchList& getSuccessors (card_t state);

  /** Prepare for computing the enabled fairness sets
   * @param state	the state to be analyzed
   * @param buf		(output) the encoded event labels (deleted by caller)
   * @param u		(output) the bit unpacker object (deleted by caller)
   * @return		the successors of the state (deleted by caller)
   */
  card_t* prepareFair (card_t state,
		       word_t*& buf,
		       class BitUnpacker*& u);

  /** Compute the enabled fairness sets
   * @param state	the state being analyzed (for diagnostics)
   * @param u		the bit unpacker object from prepareFair ()
   * @param strong	(output) the enabled strong fairness sets (optional)
   *			(static allocation; not deleted by caller)
   * @return		amount and numbers of enabled weak fairness sets
   *			(static allocation; not deleted by caller)
   */
  unsigned* computeFair (card_t state,
			 class BitUnpacker* u,
			 unsigned** strong);

  /** Evaluate the gates of a property automaton
   * @param state	the state number of the reachability graph
   * @param prop	the state number of the property automaton
   * @param ps		current state in the property automaton
   * @return		the amount and numbers of enabled successor states
   */
  const unsigned* eval (card_t state,
			unsigned prop,
			const class PropertyState& ps);
  /** Finish evaluating the gates of a property automaton
   * @param enabled	an array of enabled states returned by eval ()
   */
#ifdef EXPR_COMPILE
  void eval_done (const unsigned* enabled) const
#else
  static void eval_done (const unsigned* enabled)
#endif // EXPR_COMPILE
  {
#ifdef EXPR_COMPILE
    if (!myCompilation)
#endif // EXPR_COMPILE
      delete[] enabled;
  }

private:
  /** Add an encoded state to the state space
   * @param state	the encoded state
   * @param size	length of the encoded state in bytes
   * @return		true if the state was new
   */
  bool do_addState (const void* state,
		    size_t size);

  /** Fetch an encoded state
   * @param tail	flag: retrieve from tail of list instead of head
   * @param size	(output) length of the encoded stream
   * @return		the encoded state, or NULL if none available
   */
  word_t* do_popState (bool tail, size_t& size);

  /** Dequeue an unprocessed state
   * @param breadth	true=dequeue (FIFO, queue), false=pop (LIFO, stack)
   * @return		an unprocessed state, or NULL if all processed
   */
  class GlobalMarking* pop (bool breadth);

#ifdef EXPR_COMPILE
  /** Dequeue an unprocessed state
   * @param breadth	true=dequeue (FIFO, queue), false=pop (LIFO, stack)
   * @return		true if a state was found, or false if all processed
   */
  bool popCompiled (bool breadth);

  /** Determine whether arcs should be generated */
  bool isGraph () const { return true; }
#endif // EXPR_COMPILE

  /** Add all events from the current source state to the graph */
  void addEvents ();

  /** Report a successor state
   * @param transition	the transition fired
   * @param valuation	the binding of the transition
   * @param marking	the resulting marking
   * @param rejected	flag: is the state rejected?
   * @return		true if the state was new
   */
  bool report (const class Transition& transition,
	       const class Valuation& valuation,
	       const class GlobalMarking& marking,
	       bool rejected);

  /** Report a deadlock or an error in the current state
   * @param deadlock	flag: is this a deadlock?
   */
  void reportError (bool deadlock);

public:
#ifdef EXPR_COMPILE
  /** Report a successor state
   * @param state	the resulting encoded deflated state (mandatory)
   * @param size	length of the encoded state, in bytes
   * @param rejected	flag: is the state rejected?
   * @param hidden	flag: is the transition to the state hidden?
   * @return		true if the state was new
   */
  bool report (const void* state,
	       size_t size,
	       bool rejected,
	       bool hidden);
#endif // EXPR_COMPILE

private:
  /** Report an inconsistent successor state */
  void reject ();

private:
  /** Thresold for reporting intermediate statistics */
  const unsigned myThresold;
  /** The reachability graph being constructed */
  class Graph& myGraph;
  /** Buffer for encoding events */
  class BitPacker myEventBuf;
  /** List of unprocessed states */
  class SearchList& myStates;
  /** Number of the state being analyzed */
  card_t myStateNum;
  /** Number of the state last generated (set by do_addState) */
  card_t myTargetNum;
  /** Work area for computing enabled weak fairness sets */
  unsigned* myWeaklyFair;
  /** Work area for computing enabled strong fairness sets */
  unsigned* myStronglyFair;
  /** Flag: add also existing states to the state list */
  const bool myAddOld;
};

#endif // GRAPHREPORTER_H_
