// Collection of synchronisation states -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "SyncStates.h"
#include "Net.h"
#include "Transition.h"
#include "GlobalMarking.h"

/** @file SyncStates.C
 * Collection of synchronisation states for modular analysis
 */

SyncStates::SyncStates (const class Net& net) :
  myTransitionMap (), myStateMap (0)
{
  /** number of synchronisation labels */
  unsigned numLabels = 0;

  for (unsigned i = net.getNumTransitions (); i--; ) {
    const class Transition& t = net.getTransition (i);
    if (t.getNumChildren ())
      myTransitionMap.insert (TransitionMap::value_type (&t, numLabels++));
  }

  myStateMap = new StateMap[numLabels];
}

SyncStates::~SyncStates ()
{
  for (unsigned i = myTransitionMap.size (); i--; ) {
    StateMap& m = myStateMap[i];
    for (StateMap::iterator j = m.begin (); j != m.end (); j++) {
      StateSet& s = j->second;
      for (StateSet::iterator k = s.begin (); k != s.end (); k++)
	delete[] *k;
    }
  }

  delete[] myStateMap;
}

void
SyncStates::add (const class Transition& transition,
		 const word_t* buf,
		 word_t size)
{
  assert (!!transition.getParent ());
  assert (myTransitionMap.find (transition.getParent ()) !=
	  myTransitionMap.end ());

  word_t* state = new word_t[size + 1];
  *state = size;
  memcpy (state + 1, buf, size * sizeof *buf);

  StateMap& s =
    myStateMap[myTransitionMap.find (transition.getParent ())->second];
  std::pair<StateMap::iterator,bool> p =
    s.insert (StateMap::value_type (transition.getNet (), StateSet ()));
  std::pair<StateSet::iterator,bool> q =
    p.first->second.insert (state);
  if (!q.second)
    delete[] state;
}

/** Synchronise on a label
 * @param m		the source marking
 * @param transition	the synchronisation label
 * @param sm		the local synchronisation states
 * @param child		number of the child being checked
 * @param reporter	the interface for reporting successor states
 */
static void
sync (class GlobalMarking& m,
      const class Transition& transition,
      const SyncStates::StateMap& sm,
      unsigned child,
      class StateReporter& reporter)
{
  if (child < transition.getNumChildren ()) {
    const class Net& net = *transition.getChild (child).getNet ();
    SyncStates::StateMap::const_iterator s = sm.find (&net);
    if (s == sm.end ())
      return;
    const SyncStates::StateSet& ss = s->second;
    for (SyncStates::StateSet::const_iterator i = ss.begin ();
	 i != ss.end (); i++) {
      m.decode (net, (*i) + 1);
      sync (m, transition, sm, child + 1, reporter);
    }
  }
  else
    transition.analyze (m, reporter);
}

void
SyncStates::sync (class GlobalMarking& m,
		  const class Transition& transition,
		  class StateReporter& reporter) const
{
  assert (transition.getNumChildren () > 0);
  assert (myTransitionMap.find (&transition) != myTransitionMap.end ());
  ::sync (m, transition,
	  myStateMap[myTransitionMap.find (&transition)->second],
	  0, reporter);
}

void
SyncStates::clear ()
{
  for (unsigned i = myTransitionMap.size (); i--; ) {
    StateMap& m = myStateMap[i];
    for (StateMap::iterator j = m.begin (); j != m.end (); j++) {
      StateSet& s = j->second;
      for (StateSet::iterator k = s.begin (); k != s.end (); k++)
	delete[] *k;
      s.clear ();
    }
    m.clear ();
  }
}
