// Collection of synchronisation states -*- c++ -*-

#ifndef SYNCSTATES_H_
# define SYNCSTATES_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include <map>
# include <set>
# include "BitBuffer.h"

/** @file SyncStates.h
 * Collection of synchronisation states for modular analysis
 */

/** Collection of synchronisation states for modular analysis */
class SyncStates
{
public:
  /** Map of synchronisation labels to label numbers */
  typedef std::map<const class Transition*,unsigned> TransitionMap;
  /** Comparison operator for encoded states */
  struct ltstate
  {
    /** Less-than comparison
     * @param s1	first state to be compared
     * @param s2	second state to be compared
     * @return		true if s1 < s2
     */
    bool operator() (const word_t* s1,
		     const word_t* s2) const {
      return *s1 == *s2
	? memcmp (s1 + 1, s2 + 1, *s1 * sizeof *s1)
	: (*s1 < *s2);
    }
  };
  /** Set of encoded states */
  typedef std::set<word_t*,struct ltstate> StateSet;
  /** Map from child nets to sets of state vectors */
  typedef std::map<const class Net*,StateSet> StateMap;

  /** Constructor
   * @param net		the net whose children are being synchronised
   */
  SyncStates (const class Net& net);
private:
  /** Copy constructor */
  SyncStates (const class SyncStates& old);
  /** Assignment operator */
  class SyncStates& operator= (const class SyncStates& old);
public:
  /** Destructor */
  ~SyncStates ();

  /** Add a synchronisation state
   * @param transition	the child transition
   * @param buf		the encoded source state
   * @param size	length of the encoded state in words
   */
  void add (const class Transition& transition,
	    const word_t* buf,
	    word_t size);

  /** Synchronise on a label
   * @param m		the source marking
   * @param transition	the synchronisation label
   * @param reporter	the interface for reporting successor states
   */
  void sync (class GlobalMarking& m,
	     const class Transition& transition,
	     class StateReporter& reporter) const;

  /** Clear the synchronisation states */
  void clear ();

private:
  /** Map from synchronisation labels to index numbers */
  TransitionMap myTransitionMap;
  /** Maps from nets to state sets, indexed by synchronisation label */
  StateMap* myStateMap;
};

#endif // SYNCSTATES_H_
